<li>
    <h2 class="caret caret-down" id="section2">NLP section</h2>
    <ul class="nested active">
        {% for section in nlp_subsections %}
            {{ section }}
        {% endfor %}
    </ul>
</li>
