<li>
    <h3 class="caret" id="section3">{{ title }}</h3>
    <ul class="nested">
        <li>
            <h4 class="caret" id="section3_1">Distribution of targets</h4>
            <ul class="nested">
                <p><img src={{target_distribution}} alt="target_distribution"></p>
            </ul>
        </li>
        <li>
            <h4 class="caret" id="section3_2">Predictions vs labels difference histogram</h4>
            <ul class="nested">
                <p><img src={{error_hist}} alt="error_hist"></p>
            </ul>
        </li>
        <li>
            <h4 class="caret" id="section3_3">Predictions vs labels scatter plot</h4>
            <ul class="nested">
                <p><img src={{scatter_plot}} alt="scatter_plot"></p>
            </ul>
        </li>
    </ul>
</li>
