<li>
    <h3 class="caret" id="section3">{{ title }}</h3>
    <ul class="nested">
        <p><img src={{char_len_hist}} alt="char_len_histogram"></p>
        <p><img src={{tokens_len_hist}} alt="tokens_len_histogram"></p>
    </ul>
</li>
